package v1

import (
	"errors"
	"fmt"

	"sigs.k8s.io/kubebuilder/v3/pkg/model/resource"

	"github.com/vmware-tanzu-labs/operator-builder/pkg/utils"
)

func (c *ComponentWorkload) Validate() error {
	missingFields := []string{}

	// required fields
	if c.Name == "" {
		missingFields = append(missingFields, "name")
	}

	if c.Spec.APIGroup == "" {
		missingFields = append(missingFields, "spec.apiGroup")
	}

	if c.Spec.APIVersion == "" {
		missingFields = append(missingFields, "spec.apiVersion")
	}

	if c.Spec.APIKind == "" {
		missingFields = append(missingFields, "spec.apiKind")
	}

	if len(missingFields) > 0 {
		msg := fmt.Sprintf("Missing required fields: %s", missingFields)

		return errors.New(msg)
	}

	return nil
}

func (c *ComponentWorkload) GetWorkloadKind() WorkloadKind {
	return c.Kind
}

// methods that implement WorkloadAPIBuilder.
func (c *ComponentWorkload) GetName() string {
	return c.Name
}

func (c *ComponentWorkload) GetPackageName() string {
	return c.PackageName
}

func (c *ComponentWorkload) GetAPIGroup() string {
	return c.Spec.APIGroup
}

func (c *ComponentWorkload) GetAPIVersion() string {
	return c.Spec.APIVersion
}

func (c *ComponentWorkload) GetAPIKind() string {
	return c.Spec.APIKind
}

func (c *ComponentWorkload) GetSubcommandName() string {
	return c.Spec.CompanionCliSubcmd.Name
}

func (c *ComponentWorkload) GetSubcommandDescr() string {
	return c.Spec.CompanionCliSubcmd.Description
}

func (c *ComponentWorkload) GetSubcommandVarName() string {
	return c.Spec.CompanionCliSubcmd.VarName
}

func (c *ComponentWorkload) GetSubcommandFileName() string {
	return c.Spec.CompanionCliSubcmd.FileName
}

func (*ComponentWorkload) GetRootcommandName() string {
	// no root commands for component workloads
	return ""
}

func (c *ComponentWorkload) IsClusterScoped() bool {
	return c.Spec.ClusterScoped
}

func (*ComponentWorkload) IsStandalone() bool {
	return false
}

func (*ComponentWorkload) IsComponent() bool {
	return true
}

func (*ComponentWorkload) IsCollection() bool {
	return false
}

func (c *ComponentWorkload) SetSpecFields(workloadPath string) error {
	apiSpecFields, err := processMarkers(workloadPath, c.Spec.Resources, false)
	if err != nil {
		return err
	}

	c.Spec.APISpecFields = apiSpecFields

	return nil
}

func (c *ComponentWorkload) SetResources(workloadPath string) error {
	sourceFiles, rbacRules, ownershipRules, err := processResources(workloadPath, c.Spec.Resources)
	if err != nil {
		return err
	}

	c.Spec.SourceFiles = *sourceFiles
	c.Spec.RBACRules = *rbacRules
	c.Spec.OwnershipRules = *ownershipRules

	return nil
}

func (c *ComponentWorkload) GetDependencies() []*ComponentWorkload {
	return c.Spec.ComponentDependencies
}

func (*ComponentWorkload) SetComponents(components []*ComponentWorkload) error {
	return errors.New("cannot set component workloads on a component workload - only on collections")
}

func (c *ComponentWorkload) HasChildResources() bool {
	return len(c.Spec.Resources) > 0
}

func (*ComponentWorkload) GetComponents() []*ComponentWorkload {
	return []*ComponentWorkload{}
}

func (c *ComponentWorkload) GetSourceFiles() *[]SourceFile {
	return &c.Spec.SourceFiles
}

func (c *ComponentWorkload) GetFuncNames() (createFuncNames, initFuncNames []string) {
	return getFuncNames(*c.GetSourceFiles())
}

func (c *ComponentWorkload) GetAPISpecFields() []*APISpecField {
	return c.Spec.APISpecFields
}

func (c *ComponentWorkload) GetRBACRules() *[]RBACRule {
	return &c.Spec.RBACRules
}

func (c *ComponentWorkload) GetOwnershipRules() *[]OwnershipRule {
	return &c.Spec.OwnershipRules
}

func (c *ComponentWorkload) GetComponentResource(domain, repo string, clusterScoped bool) *resource.Resource {
	var namespaced bool
	if clusterScoped {
		namespaced = false
	} else {
		namespaced = true
	}

	api := resource.API{
		CRDVersion: "v1",
		Namespaced: namespaced,
	}

	return &resource.Resource{
		GVK: resource.GVK{
			Domain:  domain,
			Group:   c.Spec.APIGroup,
			Version: c.Spec.APIVersion,
			Kind:    c.Spec.APIKind,
		},
		Plural: utils.PluralizeKind(c.Spec.APIKind),
		Path: fmt.Sprintf(
			"%s/apis/%s/%s",
			repo,
			c.Spec.APIGroup,
			c.Spec.APIVersion,
		),
		API:        &api,
		Controller: true,
	}
}

func (c *ComponentWorkload) HasSubCmdName() bool {
	return c.Spec.CompanionCliSubcmd.Name != ""
}

func (c *ComponentWorkload) SetNames() {
	c.PackageName = utils.ToPackageName(c.Name)
	if c.HasSubCmdName() {
		c.Spec.CompanionCliSubcmd.VarName = utils.ToPascalCase(c.Spec.CompanionCliSubcmd.Name)
		c.Spec.CompanionCliSubcmd.FileName = utils.ToFileName(c.Spec.CompanionCliSubcmd.Name)
	}
}
